""" fixtures for SAX tests """

from pytest import fixture

import sax
from sax.funcs.pic import dc_transmission, dc_coupling, wg_transmission


@fixture
def dc():
    dc = sax.model(
        funcs={
            ("p0", "p1"): dc_transmission,
            ("p1", "p0"): dc_transmission,
            ("p2", "p3"): dc_transmission,
            ("p3", "p2"): dc_transmission,
            ("p0", "p2"): dc_coupling,
            ("p2", "p0"): dc_coupling,
            ("p1", "p3"): dc_coupling,
            ("p3", "p1"): dc_coupling,
        },
        params={"coupling": 0.3},
    )
    return dc


@fixture
def rdc():
    dc = sax.model(
        funcs={
            ("p0", "p1"): dc_transmission,
            ("p2", "p3"): dc_transmission,
            ("p0", "p2"): dc_coupling,
            ("p1", "p3"): dc_coupling,
        },
        params={"coupling": 0.3},
        reciprocal=True,
    )
    return dc


@fixture
def wg():
    wg = sax.model(
        funcs={
            ("in", "out"): wg_transmission,
        },
        params={
            "length": 25e-6,
            "wl": 1.55e-6,
            "wl0": 1.55e-6,
            "neff": 2.34,
            "ng": 3.4,
            "loss": 0.0,
        },
        reciprocal=True,
    )
    return wg


@fixture
def mzi(dc, wg):
    mzi = sax.circuit(
        models={
            "lft": dc,
            "top": wg,
            "rgt": dc,
            "btm": wg,
        },
        connections={
            "lft:p2": "top:in",
            "lft:p1": "btm:in",
            "top:out": "rgt:p3",
            "btm:out": "rgt:p0",
        },
        ports={
            "lft:p3": "in1",
            "lft:p0": "in0",
            "rgt:p2": "out1",
            "rgt:p1": "out0",
        },
    )
    return mzi



""" Loss functions for training neural networks """

from __future__ import annotations

import jax.numpy as jnp

from typing import Dict
from .._typing import Array


def mse(x: Array, y: Array) -> float:
    """mean squared error

    Args:
        x: the prediction
        y: the target

    Returns:
        The mean squared error between prediction and target
    """
    return ((x - y) ** 2).mean()


def huber_loss(x: Array, y: Array, delta=0.5) -> float:
    """huber loss

    The huber loss is like the mean squared error close to zero and mean
    absolute error for outliers

    Args:
        x: the prediction
        y: the target
        delta: the parameter controlling where in between MSE and MAE the huber
            loss is. δ=0 corresponds to MAE. δ=1 corresponds to MSE.

    Returns:
        the huber loss between prediction and target

    """
    return ((delta ** 2) * ((1.0 + ((x - y) / delta) ** 2) ** 0.5 - 1.0)).mean()


def l2_reg(weights: Dict[str, Array]) -> float:
    """L2 regularization loss

    Args:
        weights: the neural network weight dictionary. Per convention, the
            weights should be named `w{i}` and the biasses should be named `b{i}`.

    Returns:
        the L2 regularization loss

    """
    numel = 0
    loss = 0.0
    for w in (v for k, v in weights.items() if k[0] in ("w", "b")):
        numel = numel + w.size
        loss = loss + (jnp.abs(w) ** 2).sum()
    return loss / numel



""" Useful functions for working with SAX. """

from __future__ import annotations

from typing import (
    Any,
    Tuple,
    Dict,
)
from ._typing import (
    ComplexFloat,
    PDict,
    SDict,
    is_complex_float,
)


def copy_params(params: PDict) -> PDict:
    """copy a parameter dictionary

    Args:
        params: the parameter dictionary to copy

    Returns:
        the copied parameter dictionary

    Note:
        this copy function works recursively on all subdictionaries of the params
        dictionary but does NOT copy any non-dictionary values.
    """
    _params = {}
    for k, v in params.items():
        if isinstance(v, dict):
            _params[k] = copy_params(v)
        elif is_complex_float(v):
            _params[k] = v
        else:
            raise ValueError(
                "params dictionary to copy does not have the right type format"
            )
    return _params


def get_ports(sdict: SDict) -> Tuple[str, ...]:
    """get port names of an sdict

    Args:
        sdict: the SDict dictionary to get the port names from
    """
    ports: Dict[str, Any] = {}
    for key in sdict:
        p1, p2 = key
        ports[p1] = None
        ports[p2] = None
    return tuple(p for p in ports)


def merge_dicts(*dicts: Dict) -> Dict:
    """merge nested dictionaries

    Args:
        *dicts: the dictionaries to merge from left to right, i.e. values in
        dictionaries more to the right get precedence.

    Yields:
        the merged dictionary

    """
    if len(dicts) == 1:
        return dict(_generate_merged_dict(dicts[0], {}))
    elif len(dicts) == 2:
        return dict(_generate_merged_dict(dicts[0], dicts[1]))
    else:
        return merge_dicts(dicts[0], merge_dicts(*dicts[1:]))


def reciprocal(sdict: SDict) -> SDict:
    """Make an SDict reciprocal

    Args:
        sdict: theSDict to make reciprocal

    Returns:
        the reciprocal SDict
    """
    return {
        **{(p1, p2): v for (p1, p2), v in sdict.items()},
        **{(p2, p1): v for (p1, p2), v in sdict.items()},
    }


def rename_ports(sdict: SDict, ports: Dict[str, str]) -> SDict:
    """rename the ports of an SDict

    Args:
        sdict: the SDict to rename the ports for
        ports: a port mapping (dictionary) with keys the old names and values
            the new names.
    """
    original_ports = get_ports(sdict)
    assert len(ports) == len(original_ports)
    return {(ports[p1], ports[p2]): v for (p1, p2), v in sdict.items()}


def set_params(params: PDict, *compnames: str, **kwargs: ComplexFloat) -> PDict:
    """update a parameter dictionary

    add or update the given keyword arguments to each (sub)dictionary of the
    given params dictionary

    Args:
        params: the parameter dictionary to update with the given parameters
        *compnames: the nested component names for which to set the parameters.
            If left out, the given parameters will be applied globally to all
            (sub)components in the dictionary.
        **kwargs: the parameters to update the parameter dictionary with.

    Returns:
        The modified dictionary.

    Note:
        - Even though it's possible to update parameter dictionaries in place,
          this function is convenient to apply certain parameters (e.g.
          wavelength 'wl' or temperature 'T') globally.
        - This operation never updates the given params dictionary inplace.
        - Any non-float keyword arguments will be silently ignored.

    Example:
        This is how to change the wavelength to 1600nm for each component in
        the nested parameter dictionary::

            params = set_params(params, wl=1.6e-6)

        Or to set the temperature for only the direcional coupler named 'dc'
        belonging to the MZI named 'mzi' in the circuit::

            params = set_params(params, "mzi", "dc", T=30.0)
    """
    _params = {}
    if not compnames:
        for k, v in params.items():
            if isinstance(v, dict):
                _params[k] = set_params(v, **kwargs)
            elif is_complex_float(v):
                if k in kwargs and is_complex_float(kwargs[k]):
                    _params[k] = kwargs[k]
                else:
                    _params[k] = v
            else:
                raise ValueError(
                    "params dictionary to set global parameters for "
                    "does not have the right type format"
                )
    else:
        for k, v in params.items():
            if isinstance(v, dict):
                if k == compnames[0]:
                    _params[k] = set_params(v, *compnames[1:], **kwargs)
                else:
                    _params[k] = set_params(v)
            elif is_complex_float(v):
                _params[k] = v
            else:
                raise ValueError(
                    "params dictionary to set global parameters for "
                    "does not have the right type format"
                )
    return _params


def _generate_merged_dict(dict1: Dict, dict2: Dict) -> Dict:
    """merge two (possibly deeply nested) dictionaries

    Args:
        dict1: the first dictionary to merge
        dict2: the second dictionary to merge

    Yields:
        the merged dictionary

    """
    # from https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(_generate_merged_dict(dict1[k], dict2[k])))  # type: ignore
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                yield (k, dict2[k])
                # Alternatively, replace this with exception raiser to alert you of value conflicts
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])



""" Tests for sax.utils """

import sax

from .fixtures import dc, wg, mzi


def test_get_ports(dc):
    ports = sax.get_ports(dc)
    assert ports == ("p0", "p1", "p2", "p3")


def test_rename_ports(dc):
    portmap = {
        "p0": "in0",
        "p1": "out0",
        "p2": "out1",
        "p3": "in1",
    }
    _dc = sax.rename_ports(dc, portmap)
    ports = sax.get_ports(_dc)
    assert ports == ("in0", "out0", "out1", "in1")


def test_set_params(mzi):
    params = sax.set_params(mzi.params, wl=3.0e-6)

    def _recurse_params(params, wl=3.0e-6):
        for k, v in params.items():
            if isinstance(v, dict):
                _recurse_params(v, wl=wl)
            elif k == "wl":
                assert v == wl

    _recurse_params(params, wl=3.0e-6)
    params = sax.set_params(params, "btm", wl=5.0e-6)
    _recurse_params(params.pop("btm"), wl=5.0e-6)
    _recurse_params(params, wl=3.0e-6)



""" SAX thin-film models """

from __future__ import annotations

import jax.numpy as jnp
from ..core import model


@model(params={"ni": 1.0, "nj": 1.0, "wl": 532.0})
def fresnel_mirror_ij(params):
    """Model a (fresnel) interface between twoo refractive indices

    Args:
        ni: refractive index of the initial medium
        nf: refractive index of the final
        wl: the wavelength of the light
    """
    r_fresnel_ij = (params["ni"] - params["nj"]) / (
        params["ni"] + params["nj"]
    )  # i->j reflection
    t_fresnel_ij = 2 * params["ni"] / (params["ni"] + params["nj"])  # i->j transmission
    r_fresnel_ji = -r_fresnel_ij  # j -> i reflection
    t_fresnel_ji = (1 - r_fresnel_ij ** 2) / t_fresnel_ij  # j -> i transmission
    sdict = {
        ("in", "in"): r_fresnel_ij,
        ("in", "out"): t_fresnel_ij,
        ("out", "in"): t_fresnel_ji,
        ("out", "out"): r_fresnel_ji,
    }
    return sdict


@model(params={"ni": 1.0, "di": 500.0, "wl": 532.0})
def propagation_i(params):
    """Model the phase shift acquired as a wave propagates through medium A

    Args:
        ni: refractive index of medium (at wavelength wl)
        di: thickness of layer
        wl: wavelength
    """
    prop_i = jnp.exp(1j * 2 * jnp.pi * params["ni"] / params["wl"] * params["di"])
    sdict = {
        ("in", "out"): prop_i,
        ("out", "in"): prop_i,
    }
    return sdict


@model(params={"t_amp": jnp.sqrt(0.5), "t_ang": 0.0})
def mirror(params):
    r_complex_val = _r_complex(params["t_amp"], params["t_ang"])
    t_complex_val = _t_complex(params["t_amp"], params["t_ang"])
    sdict = {
        ("in", "in"): r_complex_val,
        ("in", "out"): t_complex_val,
        ("out", "in"): t_complex_val,  # (1 - r_complex_val**2)/t_complex_val, # t_ji
        ("out", "out"): r_complex_val,  # -r_complex_val, # r_ji
    }
    return sdict


def _t_complex(t_amp, t_ang):
    return t_amp * jnp.exp(-1j * t_ang)


def _r_complex(t_amp, t_ang):
    r_amp = jnp.sqrt((1.0 - t_amp ** 2))
    r_ang = t_ang - jnp.pi / 2
    return r_amp * jnp.exp(-1j * r_ang)



""" SAX Photonic Integrated Circuit models """

from __future__ import annotations

import jax.numpy as jnp

from ..core import model
from .._typing import PDict


@model(
    params={
        "length": 25e-6,
        "wl": 1.55e-6,
        "wl0": 1.55e-6,
        "neff": 2.34,
        "ng": 3.4,
        "loss": 0.0,
    },
)
def straight(params):
    """a simple straight waveguide model

    Args:
        wl: wavelength
        neff: waveguide effective index
        ng: waveguide group index (used for linear neff dispersion)
        wl0: center wavelength at which neff is defined
        length: [m] wavelength length
        loss: [dB/m] waveguide loss
    """
    neff = params["neff"]
    dwl = params["wl"] - params["wl0"]
    dneff_dwl = (params["ng"] - params["neff"]) / params["wl0"]
    neff = neff - dwl * dneff_dwl
    phase = 2.0 * jnp.pi * neff * params["length"] / params["wl"]
    transmission = 10 ** (-params["loss"] * params["length"] / 20) * jnp.exp(1j * phase)
    sdict = {
        ("in", "out"): transmission,
        ("out", "in"): transmission,
    }
    return sdict


@model(params={"coupling": 0.5})
def coupler(params: PDict):
    coupling = params["coupling"]
    assert not isinstance(coupling, dict)
    kappa = coupling ** 0.5
    tau = (1 - coupling) ** 0.5
    sdict = {
        ("p0", "p1"): tau,
        ("p1", "p0"): tau,
        ("p2", "p3"): tau,
        ("p3", "p2"): tau,
        ("p0", "p2"): 1j * kappa,
        ("p2", "p0"): 1j * kappa,
        ("p1", "p3"): 1j * kappa,
        ("p3", "p1"): 1j * kappa,
    }
    return sdict



""" Neural network models and utilities """

from __future__ import annotations

from . import _io as io
from . import loss
from . import nn
from . import utils

from .utils import (
    cartesian_product,
    denormalize,
    get_df_columns,
    get_normalization,
    norm,
    normalize,
)

from .nn import dense, preprocess, generate_dense_weights

from .loss import huber_loss, l2_reg, mse

from ._io import (
    get_available_sizes,
    get_dense_weights_path,
    get_norm_path,
    load_nn_dense,
    load_nn_weights_json,
    save_nn_weigths_json,
)



""" Common datastructure types used in SAX """

from __future__ import annotations

from textwrap import indent

import numpy as np
import jax.numpy as jnp

from typing import Any, NamedTuple, Dict, Union, Tuple, Callable

__all__ = ["Array", "Float", "ComplexFloat", "PDict", "Sdict", "ModelFunc", "Model"]

Array = Union[jnp.ndarray, np.ndarray]

Float = Union[float, Array]
""" a ComplexFloat type containing floats and float arrays """

ComplexFloat = Union[complex, Float]
""" a ComplexFloat type containing complex floats and complex float arrays """

PDict = Union[Dict[str, "PDict"], Dict[str, ComplexFloat]]
""" a Parameter Dictionary type """

SDict = Dict[Tuple[str, str], ComplexFloat]
""" a S-Parameter Dictionary type """

ModelFunc = Callable[[PDict], SDict]
""" a Model Function type which takes a single Parameter Dictionary as
arguments and returns an S parameter dictionary """


class Model(NamedTuple):
    """a SAX Model consists of a model function and its corresponding parameters

    Args:
        func: the model function.
        params: the default parameters belonging to the model function.

    Returns:
        the SAX model function
    """

    func: ModelFunc
    params: PDict

    def __repr__(self):
        s = f"{self.__class__.__name__}(\n"
        s += indent(f"func={self.func}", "    ")
        s += ",\n"
        s += indent(f"params={self._repr_dict(self.params)}", "    ")
        s += ",\n)"
        return s

    @staticmethod
    def _repr_dict(params) -> str:
        s = "{\n"
        for k, v in params.items():
            if isinstance(v, dict):
                s += indent(f"{repr(k)}: {Model._repr_dict(v)}", "    ") + ",\n"
            else:
                s += f"    {repr(k)}: {v},\n"
        s += "}"
        return s


def is_float(x: Any) -> bool:
    """check if an object is a float or a float array"""
    if isinstance(x, float):
        return True
    if isinstance(x, np.ndarray):
        return x.dtype in (np.float16, np.float32, np.float64, np.float128)
    if isinstance(x, jnp.ndarray):
        return x.dtype in (jnp.float16, jnp.float32, jnp.float64)
    return False


def is_complex(x: Any) -> bool:
    """check if an object is a complex or a complex array"""
    if isinstance(x, complex):
        return True
    if isinstance(x, np.ndarray):
        return x.dtype in (np.complex64, np.complex128)
    if isinstance(x, jnp.ndarray):
        return x.dtype in (jnp.complex64, jnp.complex128)
    return False


def is_complex_float(x: Any) -> bool:
    """check if an object is a complex float or a complex float array"""
    return is_float(x) or is_complex(x)



""" Tests for sax.core """

import sax
import jax
import jax.numpy as jnp
from pytest import approx

from .fixtures import dc, rdc, mzi, wg


def test_model(dc):
    assert dc.funcs["p0", "p1"] == dc.funcs["p1", "p0"]
    assert dc.funcs["p2", "p3"] == dc.funcs["p3", "p2"]
    assert dc.funcs["p0", "p2"] == dc.funcs["p2", "p0"]
    assert dc.funcs["p1", "p3"] == dc.funcs["p3", "p1"]
    assert tuple(dc.params.keys()) == ("coupling",)


def test_reciprocal_model(dc, rdc):
    assert rdc.funcs["p0", "p1"] == rdc.funcs["p1", "p0"]
    assert rdc.funcs["p2", "p3"] == rdc.funcs["p3", "p2"]
    assert rdc.funcs["p0", "p2"] == rdc.funcs["p2", "p0"]
    assert rdc.funcs["p1", "p3"] == rdc.funcs["p3", "p1"]
    assert dc.funcs["p0", "p1"] == rdc.funcs["p1", "p0"]
    assert dc.funcs["p2", "p3"] == rdc.funcs["p3", "p2"]
    assert dc.funcs["p0", "p2"] == rdc.funcs["p2", "p0"]
    assert dc.funcs["p1", "p3"] == rdc.funcs["p3", "p1"]
    assert tuple(rdc.params.keys()) == ("coupling",)


def test_circuit_mzi(mzi):
    assert tuple(sorted(mzi.params.keys())) == ("btm", "lft", "rgt", "top")
    mzi.params["lft"]["coupling"] = 0.5
    mzi.params["rgt"]["coupling"] = 0.5
    mzi.params["top"]["length"] = 50e-6
    mzi.params["btm"]["length"] = 50e-6
    assert abs(mzi.funcs["in0", "out0"](mzi.params)) == approx(0.0)
    mzi.params["btm"]["length"] = 25e-6
    assert abs(mzi.funcs["in0", "out0"](mzi.params)) == approx(0.7248724)
    f = jax.jit(mzi.funcs["in0", "out0"])
    assert abs(f(mzi.params)) == approx(0.7248724)
    assert abs(f(mzi.params)) == approx(0.7248724)
    g = jax.grad(lambda params: jnp.abs(mzi.funcs["in0", "out0"](params)) ** 2)
    assert g(mzi.params)["btm"]["length"] == approx(4736649.0)


def test_circuit_mzi_mzi(wg, mzi):
    mzi.params["lft"]["coupling"] = 0.5
    mzi.params["rgt"]["coupling"] = 0.5
    mzi.params["top"]["length"] = 50e-6
    mzi.params["btm"]["length"] = 50e-6
    mzi_mzi = sax.circuit(
        models={
            "lft": mzi,
            "top": wg,
            "rgt": mzi,
            "btm": wg,
        },
        connections={
            "lft:out1": "top:in",
            "lft:out0": "btm:in",
            "top:out": "rgt:in1",
            "btm:out": "rgt:in0",
        },
        ports={
            "lft:in1": "in1",
            "lft:in0": "in0",
            "rgt:out1": "out1",
            "rgt:out0": "out0",
        },
    )
    assert tuple(sorted(mzi_mzi.params.keys())) == ("btm", "lft", "rgt", "top")
    assert abs(mzi_mzi.funcs["in0", "out0"](mzi_mzi.params)) == approx(1.0)
    mzi_mzi.params["lft"]["btm"]["length"] = 25e-6
    mzi_mzi.params["rgt"]["btm"]["length"] = 25e-6
    mzi_mzi.params["btm"]["length"] = 25e-6
    assert abs(mzi_mzi.funcs["in0", "out0"](mzi_mzi.params)) == approx(1.0)
    f = jax.jit(mzi_mzi.funcs["in0", "out0"])
    assert abs(f(mzi_mzi.params)) == approx(1.0)
    assert abs(f(mzi_mzi.params)) == approx(1.0)
    g = jax.grad(lambda params: jnp.abs(mzi_mzi.funcs["in0", "out0"](params)) ** 2)
    assert g(mzi_mzi.params)["btm"]["length"] == approx(-0.4917065)



""" SAX core """

from __future__ import annotations

from functools import wraps, partial

import jax
import jax.numpy as jnp

from .utils import merge_dicts
from typing import (
    Dict,
    Iterable,
    Optional,
    Tuple,
    Union,
    overload,
)
from ._typing import (
    Model,
    ModelFunc,
    PDict,
    SDict,
)


def circuit(
    instances: Dict[str, Model],
    connections: Dict[str, str],
    ports: Dict[str, str],
    auto_prune: bool = False,
    keep: Optional[Iterable[Tuple[str, str]]] = None,
) -> Model:
    """generate a circuit model for the instance models given

    Args:
        instances: a dictionary with as keys the model names and values
            the model dictionaries.
        connections: a dictionary where both keys and values are strings of the
            form "modelname:portname"
        ports: a dictionary mapping portnames of the form
            "modelname:portname" to new unique portnames
        auto_prune: remove zero-valued connections and connections between non-output ports
            *while* evaluating the circuit SDict. This results in noticeably better
            performance and lower memory usage.  However, it also makes the
            resulting circuit non-jittable!
        keep: output port combinations to keep. All other combinations will be
            removed from the final sdict. Note: only output ports specified as
            *values* in the ports dict will be considered. For any port combination
            given, the reciprocal equivalent will automatically be added. This flag
            can be used in stead of ``auto_prune=True`` with jax.jit if you know in
            advance which port combinations of the sdict you're interested in.

    Returns:
        the circuit model with the given port names.

    Example:
        A simple mzi can be created as follows::

            import sax
            mzi = sax.circuit(
                instances = {
                    "lft": coupler_model,
                    "top": waveguide_model,
                    "btm": waveguide_model,
                    "rgt": coupler_model,
                },
                connections={
                    "lft:out0": "btm:in0",
                    "btm:out0": "rgt:in0",
                    "lft:out1": "top:in0",
                    "top:out0": "rgt:in1",
                },
                ports={
                    "lft:in0": "in0",
                    "lft:in1": "in1",
                    "rgt:out0": "out0",
                    "rgt:out1": "out1",
                },
            )
    """
    instances, connections, ports = _validate_circuit_parameters(
        instances, connections, ports
    )
    params = {k: v.params for k, v in instances.items()}
    funcs = {k: v.func for k, v in instances.items()}

    if keep:
        keep_dict = {min(p1, p2): max(p1, p2) for p1, p2 in keep}
        keep = tuple((p1, p2) for p1, p2 in keep_dict.items())

    def circuit(params, auto_prune=auto_prune, keep=keep):
        sdicts = {name: func(params[name]) for name, func in funcs.items()}
        return evaluate_circuit(
            sdicts, connections, ports, auto_prune=auto_prune, keep=keep
        )

    return model(circuit, params)


def evaluate_circuit(
    instances: Dict[str, SDict],
    connections: Dict[str, str],
    ports: Dict[str, str],
    auto_prune: bool = False,
    keep: Optional[Iterable[Tuple[str, str]]] = None,
):
    """evaluate a circuit for the sdicts (instances) given.

    Args:
        instances: a dictionary with as keys the instance names and values
            the corresponding SDicts.
        connections: a dictionary where both keys and values are strings of the
            form "instancename:portname"
        ports: a dictionary mapping portnames of the form
            "instancename:portname" to new unique portnames
        auto_prune: remove zero-valued connections and connections between
            non-output ports *while* evaluating the circuit SDict. This results in
            noticeably better performance and lower memory usage.  However, it also
            makes the resulting circuit non-jittable!
        keep: output port combinations to keep. All other combinations will be
            removed from the final sdict. Note: only output ports specified as
            *values* in the ports dict will be considered. For any port combination
            given, the reciprocal equivalent will automatically be added. This flag
            can be used in stead of ``auto_prune=True`` with jax.jit if you know in
            advance which port combinations of the sdict you're interested in.

    Returns:
        the circuit model dictionary with the given port names.

    Example:
        The SDict for a very simple mzi can for example be evaluated as follows::

            import sax
            S_mzi = sax.evaluate_circuit(
                instances = {
                    "lft": S_coupler,
                    "top": S_waveguide,
                    "rgt": S_coupler,
                },
                connections={
                    "lft:out0": "rgt:in0",
                    "lft:out1": "top:in0",
                    "top:out0": "rgt:in1",
                },
                ports={
                    "lft:in0": "in0",
                    "lft:in1": "in1",
                    "rgt:out0": "out0",
                    "rgt:out1": "out1",
                },
            )
    """
    float_eps = 2 * jnp.finfo(jnp.zeros(0, dtype=float).dtype).resolution

    if keep:
        keep = set(list(keep) + [(p2, p1) for p1, p2 in keep])

    block_diag = {}
    for name, sdict in instances.items():
        block_diag.update(
            {(f"{name}:{p1}", f"{name}:{p2}"): v for (p1, p2), v in sdict.items()}
        )

    sorted_connections = sorted(connections.items(), key=_connections_sort_key)
    all_connected_instances = {k: {k} for k in instances}
    for k, l in sorted_connections:
        name1, _ = k.split(":")
        name2, _ = l.split(":")

        connected_instances = (
            all_connected_instances[name1] | all_connected_instances[name2]
        )
        for name in connected_instances:
            all_connected_instances[name] = connected_instances

        current_ports = tuple(
            p
            for instance in connected_instances
            for p in set([p for p, _ in block_diag] + [p for _, p in block_diag])
            if p.startswith(f"{instance}:")
        )

        block_diag.update(_interconnect_ports(block_diag, current_ports, k, l))

        for i, j in list(block_diag.keys()):
            is_connected = i == k or i == l or j == k or j == l
            is_in_output_ports = i in ports and j in ports
            if is_connected and not is_in_output_ports:
                del block_diag[i, j]  # we're not interested in these port combinations

    if auto_prune:
        circuit_sdict = {
            (ports[i], ports[j]): v
            for (i, j), v in block_diag.items()
            if i in ports and j in ports and jnp.any(jnp.abs(v) > float_eps)
        }
    elif keep:
        circuit_sdict = {
            (ports[i], ports[j]): v
            for (i, j), v in block_diag.items()
            if i in ports and j in ports and (ports[i], ports[j]) in keep
        }
    else:
        circuit_sdict = {
            (ports[i], ports[j]): v
            for (i, j), v in block_diag.items()
            if i in ports and j in ports
        }
    return circuit_sdict


@overload
def model(func: None = None, params: Optional[PDict] = None) -> partial:
    ...


@overload
def model(func: ModelFunc = None, params: Optional[PDict] = None) -> Model:
    ...


def model(
    func: Optional[ModelFunc] = None, params: Optional[PDict] = None
) -> Union[Model, partial]:
    """create a SAX model from a model function and its parameters

    Args:
        func: the model function to create a model for.
        params: the default parameters belonging to the model function.

    Returns:
        the SAX model function
    """
    if func is None:
        return partial(model, params=params)  # type: ignore

    default_params = params if params is not None else {}

    if isinstance(func, Model):
        default_params = dict(merge_dicts(func.params, default_params))
        func = func.func

    assert func is not None

    @wraps(func)
    def new(params: PDict):
        params = dict(merge_dicts(default_params, params))  # type: ignore
        return func(params)

    return Model(func=new, params=default_params)


def _connections_sort_key(connection):
    """sort key for sorting a connection dictionary

    Args:
        connection of the form '{instancename}:{portname}'
    """
    part1, part2 = connection
    name1, _ = part1.split(":")
    name2, _ = part2.split(":")
    return (min(name1, name2), max(name1, name2))


def _interconnect_ports(block_diag, current_ports, k, l):
    """interconnect two ports in a given model

    Args:
        model: the component for which to interconnect the given ports
        k: the first port name to connect
        l: the second port name to connect

    Returns:
        the resulting interconnected component, i.e. a component with two ports
        less than the original component.

    Note:
        The interconnect algorithm is based on equation 6 in the paper below::

          Filipsson, Gunnar. "A new general computer algorithm for S-matrix calculation
          of interconnected multiports." 11th European Microwave Conference. IEEE, 1981.
    """
    current_block_diag = {}
    for i in current_ports:
        for j in current_ports:
            vij = _calculate_interconnected_value(
                vij=block_diag.get((i, j), 0.0),
                vik=block_diag.get((i, k), 0.0),
                vil=block_diag.get((i, l), 0.0),
                vkj=block_diag.get((k, j), 0.0),
                vkk=block_diag.get((k, k), 0.0),
                vkl=block_diag.get((k, l), 0.0),
                vlj=block_diag.get((l, j), 0.0),
                vlk=block_diag.get((l, k), 0.0),
                vll=block_diag.get((l, l), 0.0),
            )
            current_block_diag[i, j] = vij
    return current_block_diag


@jax.jit
def _calculate_interconnected_value(vij, vik, vil, vkj, vkk, vkl, vlj, vlk, vll):
    """Calculate an interconnected S-parameter value

    Note:
        The interconnect algorithm is based on equation 6 in the paper below::

          Filipsson, Gunnar. "A new general computer algorithm for S-matrix calculation
          of interconnected multiports." 11th European Microwave Conference. IEEE, 1981.
    """
    result = vij + (
        vkj * vil * (1 - vlk)
        + vlj * vik * (1 - vkl)
        + vkj * vll * vik
        + vlj * vkk * vil
    ) / ((1 - vkl) * (1 - vlk) - vkk * vll)
    return result


def _validate_circuit_parameters(
    models: Dict[str, Model], connections: Dict[str, str], ports: Dict[str, str]
) -> Tuple[Dict[str, Model], Dict[str, str], Dict[str, str]]:
    """validate the netlist parameters of a circuit

    Args:
        models: a dictionary with as keys the model names and values
            the model dictionaries.
        connections: a dictionary where both keys and values are strings of the
            form "modelname:portname"
        ports: a dictionary mapping portnames of the form
            "modelname:portname" to new unique portnames

    Returns:
        the validated and possibly slightly modified models, connections and
        ports dictionaries.
    """

    for name, model in models.items():
        assert isinstance(model, Model), f"'{model}' is not a valid SAX Model"

    if not isinstance(connections, dict):
        msg = f"Connections should be a str:str dict or a list of length-2 tuples."
        connections, _connections = {}, connections
        connection_ports = set()
        for conn in _connections:
            connections[conn[0]] = conn[1]
            for port in conn:
                msg = f"Duplicate port found in connections: '{port}'"
                assert port not in connection_ports, msg
                connection_ports.add(port)

    connection_ports = set()
    for connection in connections.items():
        for port in connection:
            msg = f"Connection ports should all be strings. Got: '{port}'"
            assert isinstance(port, str), msg
            msg = f"Connection ports should have format 'modelname:port'. Got: '{port}'"
            assert len(port.split(":")) == 2, msg
            name, _ = port.split(":")
            msg = f"Model '{name}' used in connection "
            msg += f"'{connection[0]}':'{connection[1]}', "
            msg += f"but '{name}' not found in models dictionary."
            assert name in models, msg
            msg = f"Duplicate port found in connections: '{port}'"
            assert port not in connection_ports, msg
            connection_ports.add(port)

    output_ports = set()
    for port, output_port in ports.items():
        msg = f"Ports keys in 'ports' should all be strings. Got: '{port}'"
        assert isinstance(port, str), msg
        msg = f"Port values in 'ports' should all be strings. Got: '{output_port}'"
        assert isinstance(output_port, str), msg
        msg = f"Port keys in 'ports' should have format 'model:port'. Got: '{port}'"
        assert len(port.split(":")) == 2, msg
        msg = f"Port values in 'ports' shouldn't contain a ':'. Got: '{output_port}'"
        assert ":" not in output_port, msg
        msg = f"Duplicate port found in ports or connections: '{port}'"
        assert port not in connection_ports, msg
        name, _ = port.split(":")
        msg = f"Model '{name}' used in output port "
        msg += f"'{port}':'{output_port}', "
        msg += f"but '{name}' not found in models dictionary."
        assert name in models, msg
        connection_ports.add(port)
        msg = f"Duplicate port found in output ports: '{output_port}'"
        assert output_port not in output_ports, msg
        output_ports.add(output_port)

    return models, connections, ports



""" Utilities for neural network models """

from __future__ import annotations

from collections import namedtuple

import pandas as pd
import jax.numpy as jnp

from typing import Tuple
from .._typing import ComplexFloat


def cartesian_product(*arrays: ComplexFloat) -> ComplexFloat:
    """calculate the n-dimensional cartesian product

    i.e. create all possible combinations of all elements in a given collection
    of arrays.

    Args:
        *arrays:  the arrays to calculate the cartesian product for

    Returns:
        the cartesian product.
    """
    ixarrays = jnp.ix_(*arrays)
    barrays = jnp.broadcast_arrays(*ixarrays)
    sarrays = jnp.stack(barrays, -1)
    assert isinstance(sarrays, jnp.ndarray)
    product = sarrays.reshape(-1, sarrays.shape[-1])
    assert isinstance(product, jnp.ndarray)
    return product


def denormalize(
    x: ComplexFloat, mean: ComplexFloat = 0.0, std: ComplexFloat = 1.0
) -> ComplexFloat:
    """denormalize an array with a given mean and standard deviation

    Args:
        x: the array to denormalize
        mean: the mean of the denormalized tensor
        std: the standard deviation of the denormalized tensor

    Returns:
        the denormalized tensor
    """
    return x * std + mean


norm = namedtuple("norm", ("mean", "std"))


def get_normalization(x: ComplexFloat):
    """Get mean and standard deviation for a given array

    Args:
        x: the array to get the normalization for

    Return:
        the mean and standard deviation
    """
    if isinstance(x, (complex, float)):
        return x, 0.0
    return norm(x.mean(0), x.std(0))


def get_df_columns(df: pd.DataFrame, *names: str) -> Tuple[ComplexFloat, ...]:
    """Get certain columns from a pandas DataFrame as jax.numpy arrays

    Args:
        df: the dataframe to get the columns from
        names: the names of the dataframe columns to get

    Returns:
        the columns of the dataframe as a namedtuple
    """

    tup = namedtuple("params", names)
    params_list = []
    for name in names:
        column_np = df[name].values
        column_jnp = jnp.array(column_np)
        assert isinstance(column_jnp, jnp.ndarray)
        params_list.append(column_jnp.ravel())
    return tup(*params_list)


def normalize(
    x: ComplexFloat, mean: ComplexFloat = 0.0, std: ComplexFloat = 1.0
) -> ComplexFloat:
    """normalize an array with a given mean and standard deviation

    Args:
        x: the array to normalize
        mean: the mean to normalize the array with
        std: the standard deviation to normalize the array with

    Returns:
        the normalized tensor
    """
    return (x - mean) / std



""" Neural network tools and architectures """

from __future__ import annotations

import jax
import jax.numpy as jnp
from .utils import (
    normalize,
    denormalize,
)
from typing import Optional, Callable, Dict, Tuple, Union
from .._typing import ComplexFloat, ComplexFloat, Array


def preprocess(*params: ComplexFloat) -> ComplexFloat:
    """preprocess parameters

    - all arguments are first casted into the same shape
    - then pairs of arguments are divided into each other to create relative arguments.
    - all arguments are then stacked into one big tensor

    Args:
        *params: the parameters to combine into a stacked tensor. Note that all
            these parameters should be broadcastable to the same shape!
    """
    x = jnp.stack(jnp.broadcast_arrays(*params), -1)
    assert isinstance(x, jnp.ndarray)
    to_concatenate = [x]
    for i in range(1, x.shape[-1]):
        _x = jnp.roll(x, shift=i, axis=-1)
        to_concatenate.append(x / _x)
        to_concatenate.append(_x / x)
    x = jnp.concatenate(to_concatenate, -1)
    assert isinstance(x, jnp.ndarray)
    return x


def dense(
    weights: Dict[str, Array],
    *params: ComplexFloat,
    x_norm: Tuple[float, float] = (0.0, 1.0),
    y_norm: Tuple[float, float] = (0.0, 1.0),
    preprocess: Callable = preprocess,
    activation: Callable = jax.nn.leaky_relu,
) -> ComplexFloat:
    """simple dense neural network

    Args:
        weights: the weights of the dense neural network in dictionary format.
            The key's of the dictionary should go from 'w0', 'b0' to 'wN', 'bN',
            with N the number of layers.
        *params: the parameters to use as input to the neural network. Note
            that all these parameters should be broadcastable to the same shape!
        x_norm: normalization constants (mean, std) for the input data
        y_norm: normalization constants (mean, std) for the output data
    """
    x_mean, x_std = x_norm
    y_mean, y_std = y_norm
    x = preprocess(*params)
    x = normalize(x, mean=x_mean, std=x_std)
    for i in range(len([w for w in weights if w.startswith("w")])):
        x = activation(x @ weights[f"w{i}"] + weights.get(f"b{i}", 0.0))
    y = denormalize(x, mean=y_mean, std=y_std)
    return y


def generate_dense_weights(
    key: Union[int, Array],
    sizes: Tuple[int, ...],
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    preprocess=preprocess,
) -> Dict[str, ComplexFloat]:
    """Generate the weights for a dense neural network

    Args:
        key: the random jax.random.PRNGKey or seed to generate the weights with
        sizes: the sizes of the dense neural network weights
        input_names: the names of the input parameters of the neural network.
            If given, the input dimension of the neural network will be derived
            from the number of features after preprocessing these names, i.e. this
            input dimension will then be prepended to `sizes`.
        output_names: the names of the output parameters the neural network
            should predict. If given, the output dimension of the neural
            network will be set to the number of output names, i.e. th number
            of output names will be appended to `sizes`.
        preprocess: the preprocessing function which will be used in the neural
            network. This preprocessing function determines the input dimension
            (if `input_names` is given.)

    Returns:
        the dictionary of random weights and biases.
    """

    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    assert isinstance(key, jnp.ndarray)

    sizes = tuple(s for s in sizes)
    if input_names:
        arr = preprocess(*jnp.ones(len(input_names)))
        assert isinstance(arr, jnp.ndarray)
        sizes = (arr.shape[-1],) + sizes
    if output_names:
        sizes = sizes + (len(output_names),)

    keys = jax.random.split(key, 2 * len(sizes))
    rand = jax.nn.initializers.lecun_normal()
    weights = {}
    for i, (m, n) in enumerate(zip(sizes[:-1], sizes[1:])):
        weights[f"w{i}"] = rand(keys[2 * i], (m, n))
        weights[f"b{i}"] = rand(
            keys[2 * i + 1],
            (
                1,
                n,
            ),
        ).ravel()

    return weights



# imports

from __future__ import annotations

import sys
import sax
import warnings
import setuptools
import subprocess

# install sax:
setuptools.setup(
    name=sax.__name__,
    version=sax.__version__,
    description=sax.__doc__,
    long_description=open("README.md").read(),
    author=sax.__author__,
    author_email="floris.laporte@gmail.com",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
    ],
)



""" SAX """

from __future__ import annotations

__author__ = "Floris Laporte"
__version__ = "0.1.0"

from ._typing import Array, Float, ComplexFloat, PDict, SDict, ModelFunc, Model

# from . import nn
from . import core
from . import utils
from . import models
from . import constants
from . import _typing as typing

from .core import circuit, evaluate_circuit, model
from .utils import (
    copy_params,
    get_ports,
    merge_dicts,
    reciprocal,
    rename_ports,
    set_params,
)



""" IO utilities for SAX neural networks """

from __future__ import annotations

import os
import re
import json

import jax.numpy as jnp


from .nn import dense, preprocess
from .utils import norm

from typing import Tuple, List, Optional, Dict, Callable
from .._typing import ComplexFloat


def load_nn_weights_json(path: str) -> Dict[str, ComplexFloat]:
    """Load json weights from given path

    Args:
        path: the path to load the json weights from

    Returns:
        a dictionary of weights
    """
    path = os.path.abspath(os.path.expanduser(path))
    weights = {}
    if os.path.exists(path):
        with open(path, "r") as file:
            for k, v in json.load(file).items():
                _v = jnp.array(v, dtype=float)
                assert isinstance(_v, jnp.ndarray)
                weights[k] = _v
    return weights


def save_nn_weigths_json(weights: Dict[str, ComplexFloat], path: str):
    """Save json weights to given path

    Args:
        weights: dictionary of weights to save as json
        path: the path to load the json weights from

    """
    path = os.path.abspath(os.path.expanduser(path))
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as file:
        _weights = {}
        for k, v in weights.items():
            v = jnp.atleast_1d(jnp.array(v))
            assert isinstance(v, jnp.ndarray)
            _weights[k] = v.tolist()
        json.dump(_weights, file)


def get_available_sizes(
    dirpath: str,
    prefix: str,
    input_names: Tuple[str, ...],
    output_names: Tuple[str, ...],
) -> List[Tuple[int, ...]]:
    """Get all available hidden sizes given filename parameters

    Args:
        prefix: the prefix of the filenames to check
        dirpath: the folder within to check for matching files
        input_names: the input feature names
        outptut_names: the output (predicted) feature names

    Returns:
        the possible hidden sizes

    Note:
        this function does NOT return the input size and the output size of the
        neural network. ONLY the hidden sizes are reported. The input and
        output sizes can easily be derived from `input_names` (after
        preprocessing) and `output_names`.
    """
    all_weightfiles = os.listdir(dirpath)
    possible_weightfiles = (
        s for s in all_weightfiles if s.endswith(f"-{'-'.join(output_names)}.json")
    )
    possible_weightfiles = (
        s
        for s in possible_weightfiles
        if s.startswith(f"{prefix}-{'-'.join(input_names)}")
    )
    possible_weightfiles = (re.sub("[^0-9x]", "", s) for s in possible_weightfiles)
    possible_weightfiles = (re.sub("^x*", "", s) for s in possible_weightfiles)
    possible_weightfiles = (re.sub("x[^0-9]*$", "", s) for s in possible_weightfiles)
    possible_hidden_sizes = (s.strip() for s in possible_weightfiles if s.strip())
    possible_hidden_sizes = (
        tuple(hs.strip() for hs in s.split("x") if hs.strip())
        for s in possible_hidden_sizes
    )
    possible_hidden_sizes = (
        tuple(int(hs) for hs in s[1:-1]) for s in possible_hidden_sizes if len(s) > 2
    )
    possible_hidden_sizes = sorted(
        possible_hidden_sizes, key=lambda hs: (len(hs), max(hs))
    )
    return possible_hidden_sizes


def get_dense_weights_path(
    *sizes: int,
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    dirpath: str = "weights",
    prefix: str = "dense",
    preprocess=preprocess,
):
    """Create the SAX conventional path for a given dictionary

    Args:
        *sizes: the sizes of the dense neural network weights. This can be left
            out for a single layer neural network when specifying both
            `input_names` and `output_names`.
        input_names: the names of the input parameters of the neural network.
            If given, the input dimension of the neural network will be derived
            from the number of features after preprocessing these names, i.e. this
            input dimension will then be prepended to `sizes`.
        output_names: the names of the output parameters the neural network
            should predict. If given, the output dimension of the neural
            network will be set to the number of output names, i.e. th number
            of output names will be appended to `sizes`.
        dirpath: the folder to save the weights to
        prefix: the prefix to give the weights filename
        preprocess: the preprocessing function which will be used in the neural
            network. This preprocessing function determines the input dimension
            (if `input_names` is given.)
    """
    if input_names:
        num_inputs = preprocess(*jnp.ones(len(input_names))).shape[0]
        sizes = (num_inputs,) + sizes
    if output_names:
        sizes = sizes + (len(output_names),)
    path = os.path.abspath(os.path.join(dirpath, prefix))
    if input_names:
        path = f"{path}-{'-'.join(input_names)}"
    if sizes:
        path = f"{path}-{'x'.join(str(s) for s in sizes)}"
    if output_names:
        path = f"{path}-{'-'.join(output_names)}"
    return f"{path}.json"


def get_norm_path(
    *shape: int,
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    dirpath: str = "norms",
    prefix: str = "norm",
    preprocess=preprocess,
):
    """Create the SAX conventional path for the normalization constants

    Args:
        *shape: the shape of the norm to create the path for. This can be left
            out if `input_names` or `output_names` is specified.
        input_names: the names of the input parameters of the neural network.
            If given, the shape of the norm will be prepended with the number
            of features after preprocessing these names. (mutually exclusive
            with `output_names`)
        output_names: the names of the output parameters the neural network
            should predict. If given, the shape of the norm will be appended
            with the number of output names. (mutually exclusive with
            `input_names`)
        dirpath: the folder to save the norms to
        prefix: the prefix to give the norm filename
        preprocess: the preprocessing function which will be used in the neural
            network. This preprocessing function determines the input dimension
            (if `input_names` is given.)
    """
    if input_names and output_names:
        raise ValueError(
            "To get the norm name, one can only specify `input_names` OR `output_names`."
        )
    if input_names:
        num_inputs = preprocess(*jnp.ones(len(input_names))).shape[0]
        shape = (num_inputs,) + shape
    if output_names:
        shape = shape + (len(output_names),)
    path = os.path.abspath(os.path.join(dirpath, prefix))
    if input_names:
        path = f"{path}-{'-'.join(input_names)}"
    if shape:
        path = f"{path}-{'x'.join(str(s) for s in shape)}"
    if output_names:
        path = f"{path}-{'-'.join(output_names)}"
    return f"{path}.json"


def load_nn_dense(
    *sizes: int,
    input_names: Optional[Tuple[str, ...]] = None,
    output_names: Optional[Tuple[str, ...]] = None,
    weightprefix="dense",
    weightdirpath="weights",
    normdirpath="norms",
    normprefix="norm",
    preprocess=preprocess,
) -> Callable:
    """Load a pre-trained dense model

    Args:
        *sizes: the sizes of the dense neural network weights. This can be left
            out for a single layer neural network when specifying both
            `input_names` and `output_names`.
        input_names: the names of the input parameters of the neural network.
            If given, the input dimension of the neural network will be derived
            from the number of features after preprocessing these names, i.e. this
            input dimension will then be prepended to `sizes`.
        output_names: the names of the output parameters the neural network
            should predict. If given, the output dimension of the neural
            network will be set to the number of output names, i.e. th number
            of output names will be appended to `sizes`.
        weightdirpath: the folder to save the weights to
        weightprefix: the prefix to give the weights filename
        normdirpath: the folder to save the norms to
        normprefix: the prefix to give the norm filename
        preprocess: the preprocessing function which will be used in the neural
            network. This preprocessing function determines the input dimension
            (if `input_names` is given.)
    """
    weights_path = get_dense_weights_path(
        *sizes,
        input_names=input_names,
        output_names=output_names,
        prefix=weightprefix,
        dirpath=weightdirpath,
        preprocess=preprocess,
    )
    if not os.path.exists(weights_path):
        raise ValueError("Cannot find weights path for given parameters")
    x_norm_path = get_norm_path(
        input_names=input_names,
        prefix=normprefix,
        dirpath=normdirpath,
        preprocess=preprocess,
    )
    if not os.path.exists(x_norm_path):
        raise ValueError("Cannot find normalization for input parameters")
    y_norm_path = get_norm_path(
        output_names=output_names,
        prefix=normprefix,
        dirpath=normdirpath,
        preprocess=preprocess,
    )
    if not os.path.exists(x_norm_path):
        raise ValueError("Cannot find normalization for output parameters")
    weights = load_nn_weights_json(weights_path)
    x_norm_dict = load_nn_weights_json(x_norm_path)
    y_norm_dict = load_nn_weights_json(y_norm_path)
    x_norm = norm(x_norm_dict["mean"], x_norm_dict["std"])
    y_norm = norm(y_norm_dict["mean"], y_norm_dict["std"])
    partial_dense = _PartialDense(weights, x_norm, y_norm, input_names, output_names)
    return partial_dense


class _PartialDense:
    def __init__(self, weights, x_norm, y_norm, input_names, output_names):
        self.weights = weights
        self.x_norm = x_norm
        self.y_norm = y_norm
        self.input_names = input_names
        self.output_names = output_names

    def __call__(self, *params: ComplexFloat) -> ComplexFloat:
        return dense(self.weights, *params, x_norm=self.x_norm, y_norm=self.y_norm)

    def __repr__(self):
        return f"{self.__class__.__name__}{repr(self.input_names)}->{repr(self.output_names)}"
